(redef+)

(mutual-recursion

(defun obviously-equiv-terms (x y iff-flg)

; Warning: It is desirable to keep this reasonably in sync with untranslate1,
; specifically, giving similar attention in both to functions like implies,
; iff, and not, which depend only on the propositional equivalence class of
; each argument.

; Here we code a restricted version of equivalence of x and y, for use in
; chk-defabsstobj-method-lemmas or other places where we expect this to be
; sufficient.  The only requirement is that if (obviously-equiv-terms x y
; iff-flg), then (equal x y) a theorem (in every theory extending the
; ground-zero theory) unless iff-flg is true, in which case (iff x y) is a
; theorem.

  (or (equal x y) ; common case
      (cond ((or (variablep x)
                 (variablep y))
             nil)
            ((or (fquotep x)
                 (fquotep y))
             (and iff-flg
                  (equal (equal x *nil*)
                         (equal y *nil*))))
            ((flambda-applicationp x)
             (and (flambda-applicationp y)

; There are (at least) two ways that x and y can be obviously equivalent.

; (1) The arguments agree, and their lambdas (function symbols) are equivalent
;     but have different formals and correspondingly different bodies, for
;     example:
;       ((lambda (x y) (cons x y)) '3 '4)
;       ((lambda (u v) (cons u v)) '3 '4)

; (2) The formals in the lambdas have been permuted and the arguments have been
;     correspondingly permuted, and the bodies of the lambdas are the same, for
;     example:
;       ((lambda (x y) (cons x y)) '3 '4)
;       ((lambda (y x) (cons x y)) '4 '3)

; Of course the function symbols of x and y can be equal, which fits into both
; (1) and (2).  And the discrepancies of (1) and (2) can happen together, as in
; the following example:
;       ((lambda (x y) (cons x y)) '3 '4)
;       ((lambda (u v) (cons v u)) '4 '3)

; But it is more complicated to handle this combination in full generality, so
; we content ourselves with (1) and (2).

                  (let ((x-fn (ffn-symb x))
                        (y-fn (ffn-symb y))
                        (x-args (fargs x))
                        (y-args (fargs y)))
                    (cond
                     ((equal x-fn y-fn) ; simple case
                      (obviously-equiv-terms-lst x-args y-args))
                     (t
                      (let ((x-formals (lambda-formals x-fn))
                            (x-body (lambda-body x-fn))
                            (y-formals (lambda-formals y-fn))
                            (y-body (lambda-body y-fn)))
                        (and (eql (length x-formals) (length y-formals))
                             (or

; (1) -- see above

                              (and (obviously-equiv-terms
                                    (subcor-var x-formals y-formals x-body)
                                    y-body
                                    iff-flg)
                                   (obviously-equiv-terms-lst x-args y-args))

; (2) -- see above

                              (and (obviously-equiv-terms
                                    x-body y-body iff-flg)
                                   (obviously-equal-lambda-args
                                    x-formals (fargs x)
                                    y-formals (fargs y)))))))))))
            ((not (eq (ffn-symb x) (ffn-symb y)))
             nil)
            ((member-eq (ffn-symb x) '(implies iff))
             (and (obviously-equiv-terms (fargn x 1) (fargn y 1) t)
                  (obviously-equiv-terms (fargn x 2) (fargn y 2) t)))
            ((eq (ffn-symb x) 'not)
             (obviously-equiv-terms (fargn x 1) (fargn y 1) t))
            ((eq (ffn-symb x) 'if)
             (and (obviously-equiv-terms (fargn x 1) (fargn y 1) t)
                  (obviously-equiv-terms (fargn x 3) (fargn y 3) iff-flg)
                  (or (obviously-equiv-terms (fargn x 2) (fargn y 2) iff-flg)

; Handle case that a term is of the form (or u v).

                      (and iff-flg
                           (cond ((equal (fargn x 2) *t*)
                                  (obviously-equiv-terms
                                   (fargn y 2) (fargn y 1) t))
                                 ((equal (fargn y 2) *t*)
                                  (obviously-equiv-terms
                                   (fargn x 2) (fargn x 1) t))
                                 (t nil))))))
            (t (and (equal (length (fargs x))
                           (length (fargs y)))
                    (obviously-equiv-terms-lst (fargs x) (fargs y)))))))

(defun obviously-equiv-terms-lst (x y)

; X and y are true-lists of the same length.

  (cond ((endp x) t)
        (t (and (obviously-equiv-terms (car x) (car y) nil)
                (obviously-equiv-terms-lst (cdr x) (cdr y))))))

(defun obviously-equal-lambda-args (x-formals-tail x-args-tail y-formals
                                                   y-args)

; We know that y-formals is a permutation of x-formals.  We recur through
; x-formals and x-args, checking that correspond arguments are equal.

  (cond ((endp x-formals-tail) t)
        (t (let ((posn (position-eq (car x-formals-tail) y-formals)))
             (assert$
              posn
              (and (equal (car x-args-tail)
                          (nth posn y-args))
                   (obviously-equal-lambda-args (cdr x-formals-tail)
                                                (cdr x-args-tail)
                                                y-formals y-args)))))))

)

(defun obviously-iff-equiv-terms (x y)
  (obviously-equiv-terms x y t))

(redef-)

(i-am-here)

(defxdoc note-7-4

; Eliminate extra code for backward compatibility, now obsolete, in functions
; book-hash and cert-hash.

; Improved error messages from exceeding the call depth limit, in response to
; helpful feedback from Eric Smith.

; Here is a brief explanation of the soundness bug in functional instantiation,
; and its fix.  There was a "draconian" restriction until late 2006, which made
; it illegal for free variables in a functional subsitution to occur either in
; (a) the functionally-instantiated theorem or (b) the relevant
; functionally-instantiated constraints.  That restriction was eliminated in
; Version 3.0.2 (released December 2006); this was completely appropriate for,
; and motivated by, the desire to avoid the restriction for (a).  However, it
; was unsound to do this for (b): one is really functionally instantiating the
; universal closure of the constraints, so for example if function f is
; replaced by (lambda (u) (g u y)) in a constraint formula, say for example (p
; (f x) y), then logically we are substituting into the constraint (forall (x
; y) (p (f x) y)) to get the formula (forall (x y) (p (g x y) y)), which sadly
; is not a functional instance of the original constraint.  The new
; implementation automatically performs alpha renaming of the constraint to
; avoid such capture: in our example, it functionally instantiates the renamed
; constraint (forall (x y-renamed0) (p (f x) y-renamed0)) to obtain a suitable
; proof obligation of (p (g x y) y-renamed0).  A "[Capture]" warning is printed
; when this happens.

; Disabled executable-counterpart of function print-call-history; so now we
; don't get distracting output, for example from
; (thm (equal (print-call-history) 3)).

; Added a long comment, "Essay on Admitting a Model for Apply$ and the
; Functions that Use It", near the end of ACL2 source file apply-raw.lisp.
; This Essay is a step towards incorporating apply$, which supports a weak form
; of second-order programming and reasoning, into part of ACL2.  As mentioned
; there, [new] community books directory books/projects/apply-model/ contains
; examples that illustrate the Essay.

  :parents (release-notes)
  :short "ACL2 Version  7.4 (xx, 20xx) Notes"
  :long "<p>NOTE!  New users can ignore these release notes, because the @(see
 documentation) has been updated to reflect all changes that are recorded
 here.</p>

 <p>Below we roughly organize the changes to ACL2 since Version 7.3 into the
 following categories of changes: existing features, new features, heuristic
 and efficiency improvements, bug fixes, changes at the system level, Emacs
 support, and experimental versions.  Each change is described in just one
 category, though of course many changes could be placed in more than one
 category.</p>

 <p>Note that only ACL2 system changes are listed below.  Changes to the @(see
 books) can be found by browsing the <a
 href='https://github.com/acl2/acl2/'>ACL2+Books GitHub repository</a>, in
 particular, the raw <a
 href='https://github.com/acl2/acl2/commits/master'>commit log</a>.  Also note
 that with each release, some built-in functions that were formerly in
 @(':')@(tsee program) mode are now @('see guard')-verified @(':')@(tsee logic)
 mode functions.</p>

 <h3>Changes to Existing Features</h3>

 <p>The utility @(tsee read-file-into-string) is now a macro, not a function.
 It no longer takes @('state') as an argument (though its expansion reads the
 state), and it has new keyword arguments that allow one to read successive
 segments of the file.  See @(see read-file-into-string).</p>

 <p>The test has been slightly relaxed for determining whether any supporting
 events are missing for @(tsee defabsstobj).  Thanks to Sol Swords for sending
 an example to illustrate the issue, together with analysis of the problem.</p>

 <h3>New Features</h3>

 <h3>Heuristic and Efficiency Improvements</h3>

 <p>We now avoid the @(see invariant-risk) slowdown in the case of a
 @(':')@(tsee logic) mode function with guard @('t').  For relevant examples
 see @(see program-wrapper).</p>

 <p>A technical change in the handling of @(see type-set) computations can, in
 rare cases, result in better built-in @(see type-prescription) rules for
 functions.</p>

 <h3>Bug Fixes</h3>

 <p>Fixed a soundness bug, introduced about 10 years ago, that could occur with
 @(':use') or @(':by') @(see lemma-instance)s when a variable occurs free in a
 lambda of a functional substitution.  For an explanation see a comment in
 @('(defxdoc note-7-4 ...)') in community book
 @('books/system/doc/acl2-doc.lisp'); also, a proof of @('nil') (in Version
 7.3) that exploits this bug can be found in a comment in the definition of
 function @('remove-capture-in-constraint-lst') in the ACL2 source code.  This
 may be the first time that a proved ``theorem'' in the @(see community-books)
 has been discovered to be a non-theorem (@('tr-to-array-of-array-to-tr') in
 @('books/projects/legacy-defrstobj/typed-records.lisp')).  Thanks to Sol
 Swords for pointing out that this directory of books isn't used elsewhere; for
 now we have excluded it from the regression by adding a @('cert_pl_exclude')
 file.  Also thanks to Ruben Gamboa for fixing proofs in the @('nonstd/') books
 that failed after the bug fix.</p>

 <p>Fixed a bug in @(see proof-builder) commands @('x') and @('expand'), which
 usually gave a misleading error message when attempting to expand a
 constrained function.</p>

 <p>Fixed bugs in the tracking and disabling of @(see compound-recognizer)
 rules.  Thanks to Dmitry Nadezhin for pointing out this bug, indeed, sending
 an example and analyzing it to point to system function
 @('term-and-typ-to-lookup'), which was indeed the one that had the bug.</p>

 <p>Avoided a rare error in @(tsee hons-wash) that has happened when there is a
 second garbage collection after @('hons-wash') has commenced.  Thanks to
 Keshav Kini for telling us about the problem and providing a patch, which we
 used almost verbatim.</p>

 <h3>Changes at the System Level</h3>

 <p>An ACL2 customization file is no longer loaded by default (when it exists)
 when invoking `@('make')' from the top-level ACL2 directory.  This change is
 not necessary for `@('make regression')' because the build system for the
 books already takes care of it.  However we need this change for `@('make
 DOC')', at the least; thanks to Eric Smith for bringing our attention to a
 failure of `@('make DOC')'.</p>

 <p>(Really a books change, but affects building the manual:) By default, it is
 now an error if when saving the ACL2+books or ACL2-only manuals (web-based or
 text-based), a syntax error (marked by ``xdoc error'') is encountered.  See
 @(tsee xdoc::save) and @(tsee xdoc::save-rendered): both now have an error
 argument with default @('nil') so that by default, those new errors are not
 signaled; but in accordance with the change above, that new error argument is
 @('t') for the standard ways of building the ACL2+books and ACL2-only manuals.
 Also, two arguments have been eliminated for @('xdoc::save-rendered').</p>

 <p>An invocation of `@('make update')' to build an ACL2 executable initiates a
 build only when necessary (that is, when at least one source file or
 supporting file, like @('GNUmakefile'), has changed).  Thanks to Eric Smith
 for the suggestion.  Removed the following `@('make')' targets, which were
 suspect after this change (the first definitely no longer worked) and are
 unnecessary: @('saved_acl2'), @('saved_acl2p'), @('saved_acl2r'), and
 @('saved_acl2pr').  WARNING: @('make update') should be used with care, since
 it does not rebuild the desired executable when it already exists and is more
 recent than the sources.  For example, if you change Lisp implementations
 without changing the @('PREFIX') variable &mdash; perhaps even only changing
 the version of your Lisp &mdash; then use @('make') or @('make large'), not
 @('make update').</p>

 <p>New `@('make')' target @('clean-all') removes all generated files in the
 main ACL2 directory and its @('doc/') subdirectory, while existing target
 @('clean'), which has been renamed @('clean-lite'), cleans all of those except
 files @('*saved_acl2*') (and @('doc.lisp.backup')), essentially as before.
 Thanks to Eric Smith for suggestions leading to this change, and to Keshav
 Kini for suggesting consideration of target @('distclean') as an alias for
 @('clean-all'), which it now is.  The target @('clean') is deprecated but,
 consistent with prior usage, acts as @('clean-lite') for now.</p>

 <p>A successful build using `@('make')' shows only four lines for @(''fgrep -i
 compiler make.log')' instead of five.</p>

 <p>(GCL only) Installed a patch from Camm Maguire (GCL maintainer) to fix a
 GCL compiler bug.  Also turned off SGC to avoid an ``mprotect failure'' error
 (community book @('books/centaur/fty/tests/deftagsum-scale.lisp')).</p>

 <p>Fixed a bug that was causing an assertion failure, instead of a clean
 error, when attempting to apply @(':')@(tsee monitor) to undefined primitives,
 e.g., @(':monitor cons t').  Thanks for Keshav Kini for reporting this bug and
 sending us an initial fix.</p>

 <p>Fixed a bug in the @(see break-rewrite) commands @(':go!') and @(':go$'),
 which was causing an error message to be printed.</p>

 <h3>EMACS Support</h3>

 <p>Made minor @(see acl2-doc) improvements: fixed a bug in `@('u')' command's
 error message when already at the top topic; fixed a bug that was displaying
 the \"Parent list:\" field in reverse order; and eliminated the warning when
 loading a large manual (combined manual is at about 70M bytes now; new
 threshold is 80M bytes).</p>

 <p>Improved @(see acl2-doc) to work with other manuals; see @(see acl2-doc),
 specifically Section ``Selecting a Manual''.  Note that the command for
 initially bringing up the ACL2 User's Manual, instead of the ACL2+Books
 manual, has changed from @('(defvar *acl2-doc-top-default* 'TOP)') to
 @('(defvar *acl2-doc-manual-name* 'acl2-only)').</p>

 <p>Added new @(see acl2-doc) commands, `@('/')' and `@('W')', to go to an ACL2
 definition.  By default the search is through both the ACL2 sources and books.
 The search can be continued even outside acl2-doc by using a numeric prefix
 argument with @('control-t /'), which is just the `@('/')' command made
 available outside acl2-doc.  These two commands differ only in the defaults
 presented in the minibuffer.  The default for `@('/')' is based on the text at
 the cursor.  (This is much like the similar built-in Emacs command,
 @('meta-.'); indeed, `@('/')' is next to `@('.')' on standard keyboards.)  For
 `@('W')', the default is based on the page's topic name (as shown by the
 `@('w')' command).  See @(see acl2-doc) for details.</p>

 <p>Installed a patch for @('control-t control-e') from Keshav Kini, so that
 when there are more than two windows and one has the ACL2 shell buffer
 showing, then that one receives the form.  Consider for example windows A, B,
 and C, where the cursor is showing in A and the shell buffer is showing in C.
 Previously, the shell buffer could be placed in B to receive the form.  Now,
 the form will go into C.</p>

 <p>The @(':evisc-tuple') keyword for @(tsee trace$) was being ignored when
 @(':native t') was also supplied.  Since the use of @(':native t') sends the
 tracing command to raw Lisp, this wasn't really a bug, since no raw Lisp we
 know of supports the @(':evisc-tuple') keyword.  However, we have improved our
 custom modifications of the native @('trace') for host Lisps CCL and Allegro
 CL so that @(':evisc-tuple') is no longer ignored for these Lisps.</p>

 <h3>Experimental Versions</h3>

 <p>A new rewrite rule for ACL2(r), @('realp-implies-acl2-numberp'), is
 analogous to the existing rule @('rationalp-implies-acl2-numberp').  Thanks to
 Dmitry Nadezhin for suggesting the addition of this rule, which he observed is
 necessary for some ACL2(r) proofs.</p>

 ")
