;;; Current github version before change:
;;; 4de9bd53bc28fc4ca3db90666fa086ecf8ece94c
;;; The original functions are saved under "OLD definitions:", below.

#||
Gmail	Matt Kaufmann <matthew.j.kaufmann@gmail.com>
another stobj-let question
Sol Swords <sswords@zmail0.centtech.com>	Fri, Apr 14, 2017 at 11:40 AM
To: kaufmann@cs.utexas.edu

Hi Matt --

First off, I found what I *think* is a bug (or at least something fishy) while investigating the following problem I ran into.  For nested stobjs, the updaters for child stobjs have stobjs-in properties that don't match their formals.  For the example below, e.g.,

The following events generate a guard proof obligation that surprised me:

(defstobj myarr (myarray-field :type (array (integer 0 *) (0)) :initially 0 :resizable t))

(defstobj myparent (parent-arr :type myarr))

(defun myparent-resize-array (size myparent)
  (declare (xargs :stobjs myparent :guard (natp size)
                  :guard-hints (("goal" :do-not-induct t))))
  (stobj-let ((myarr (parent-arr myparent)))
             (myarr)
             (resize-myarray-field size myarr)
             myparent))

[Edit from Matt: actually submit this instead:
(defun myparent-resize-array (size myparent)
  (declare (xargs :stobjs myparent :guard (natp size)
                  :guard-hints
                  (("goal" :do-not-induct t
                    :in-theory
                    (disable length nth true-listp myarrp resize-myarray-field)
                    :expand ((MYARRP (NTH *PARENT-ARR* MYPARENT)))))))
  (stobj-let ((myarr (parent-arr myparent)))
             (myarr)
             (resize-myarray-field size myarr)
             myparent))

The proof obligation:
Goal'
(IMPLIES (AND (INTEGERP SIZE)
              (<= 0 SIZE)
              (TRUE-LISTP MYPARENT)
              (EQUAL (LENGTH MYPARENT) 1)
              (TRUE-LISTP (NTH *PARENT-ARR* MYPARENT))
              (EQUAL (LENGTH (NTH *PARENT-ARR* MYPARENT))
                     1)
              (MYARRAY-FIELDP (NTH 0 (NTH *PARENT-ARR* MYPARENT))))
         (MYARRP (RESIZE-MYARRAY-FIELD SIZE (NTH *PARENT-ARR* MYPARENT)))).


I think this is basically checking that the new value of the stobj field to be installed in the parent stobj is still well-formed.  This surprised me because usually stobj recognizers can be just assumed without needing to be proved, because the syntactic checking ensures that all possible executions preserve the stobj invariants.  Is it necessary for this to be checked in the case of updates of child stobjs by stobj-let?

One reason I suspect this doesn't need to be checked is that I can fix it in a way that has more or less no effect on what is executed, as follows:

(defun myarr-fix (myarr)
  (declare (xargs :stobjs myarr))
  (mbe :logic (non-exec (if (myarrp myarr) myarr (create-myarr)))
       :exec myarr))

(defun myparent-resize-array (size myparent)
  (declare (xargs :stobjs myparent :guard (natp size)
                  :guard-hints (("goal" :do-not-induct t))))
  (stobj-let ((myarr (parent-arr myparent)))
             (myarr)
             (let* ((myarr (resize-myarray-field size myarr)))
               (myarr-fix myarr))
             myparent))

This makes it so the update, whose myarrp guard is checked, is done with the myarr-fix (which we know produces a myarrp), but myarr-fix's myarrp guard isn't checked.  Unfortunately, adding the myarr-fix messes up the logical meaning of the function in an ugly way so I'd like to avoid that.

I think the cause of the problem is that UPDATE-PARENT-ARR has formals (V MYPARENT) but has stobjs-in = (MYARR MYPARENT), so optimize-stobj-recognizers doesn't strip out the call of (myarrp v).  That seems fishy to me -- aren't each of the stobjs-in supposed to be either NIL or the name of the formal (when it's a stobj)?

- Sol
||#

;;; First I re-read :doc stobj-let and ran Sol's events.
;;; I obsessed too long on trying to get the exact output that he got.

;;; Then I simply tried this, before thinking it through (which I'd do if this
;;; works):

;;; meta-. defstobj-fn
;;; [irrelevant: meta-. chk-acceptable-defstobj]
;;; meta-. defstobj-axiomatic-defs [from the process-embedded-events call in
;;;   defstobj-fn]
;;; meta-. defstobj-field-fns-axiomatic-defs
;;; Oops, maybe it's the field templates.
;;; (trace$ defstobj-template) [inspecting defstobj-fn]
;;; :u
;;; (defstobj myparent (parent-arr :type myarr))
;;; Nope, no "V"
;;; Search for 'v in the file other-events.lisp
;;;   -- ouch, found in chk-acceptable-defstobj
;;;      -- but I can worry about that later (thus mark with !! here -- DONE,
;;;         and I also searched for " v ")
;;; (trace$ defstobj-field-fns-axiomatic-defs)
;;; :u
;;; (defstobj myparent (parent-arr :type myarr))
;;; Hmmm, where does "V" come from in the formals?
;;; Aha, in updater-name binding in defstobj-field-fns-axiomatic-defs.
;;; (untrace$)
;;; (redef+)
;;; Then submit this, followed by (redef-) -- but first start over, to make
;;; sure we get the same output from Sol's example either way.
;;; This definition is equivalent to the original -- we can run compare-windows
;;; against what I began with, to see the introduction of v-formal, in
;;; anticipation of binding it to something other than 'V.

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (v-formal 'v)
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i ,v-formal ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t))
                             (update-nth-array ,n i ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (eq type-term t)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t))
                            (update-nth ,n ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; Now change the binding of v-formal as per Sol's suggestion, so that the
;;; formals of update-parent-arr are (MYARR MYPARENT), i.e., v-formal is
;;; MYARR.  From the trace of DEFSTOBJ-FIELD-FNS-AXIOMATIC-DEFS I see that the
;;; template has MYARR in it.
;;; meta-. defstobj-field-template
;;; AHA -- it's the binding of TYPE, below.
;;; And looking at the code below, I see that translate-stobj-type-to-guard
;;; will give me a clue.
;;; meta-. translate-stobj-type-to-guard
;;; I'll just use the same binding below as found for stobj-recog in
;;; translate-stobj-type-to-guard, but maybe share those later (leaving !! to
;;; remember to do that -- DONE (now using translate-declaration-to-guard).
;;; So now use :redef+/:redef- again to introduce this in a new session:

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (stobj-type ; !! share with translate-stobj-type-to-guard [DONE]
              (and (not (eq type 'state))
                   (cadr (stobjp type t wrld))))
             (v-formal (or stobj-type 'v))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i ,v-formal ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t))
                             (update-nth-array ,n i ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (eq type-term t)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t))
                            (update-nth ,n ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; Didn't help!  Using
;;; :props update-parent-arr
;;; I realized that the formals are (MYARRP MYPARENT), not (MYARR MYPARENT).
;;; OH -- the formal should be the stobj itself.

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (stobj-formal ; !! maybe share with translate-stobj-type-to-guard [DONE]
              (and (not (eq type 'state))
                   (stobjp type t wrld)
                   type))
             (v-formal (or stobj-formal 'v))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i ,v-formal ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t))
                             (update-nth-array ,n i ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (eq type-term t)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t))
                            (update-nth ,n ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; Ouch:

#||
ACL2 !>(defstobj myparent (parent-arr :type myarr))


ACL2 Error in ( DEFUN UPDATE-PARENT-ARR ...):  The form MYARR is being
used, as an argument to a call of MYARRP, where the single-threaded
object of that name is required.  But in the current context, there
are no declared stobj names.  Note:  this error occurred in the context
(MYARRP MYARR).
||#

;;; (break-on-error)
;;; (defstobj myparent (parent-arr :type myarr))
;;; :b
;;; (:form 4) ; to see defuns-fn call
;;; meta-. defuns-fn ; to see if wrld is a formal before printing
;;; (pprint *)

;;; OK, so I need a stobj declaration for the child.

;;; So restart ACL2 and submit this.  (Confession: I used :stobj instead of
;;; :stobjs and got an error, initially.)

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (stobj-formal ; !! maybe share with translate-stobj-type-to-guard [DONE]
              (and (not (eq type 'state))
                   (stobjp type t wrld)
                   type))
             (v-formal (or stobj-formal 'v))
             (stobj-xargs (and stobj-formal
                               `(:stobjs ,stobj-formal)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i ,v-formal ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t
                                             ,@stobj-xargs))
                             (update-nth-array ,n i ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (eq type-term t)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t
                                            ,@stobj-xargs))
                            (update-nth ,n ,v-formal ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

#||
ACL2 !>(defstobj myparent (parent-arr :type myarr))


ACL2 Error in ( DEFUN UPDATE-PARENT-ARR ...):  A single-threaded object,
namely MYARR, is being used where an ordinary object is expected.
||#

;;; (break-on-error)
;;; (:form 4)
;;; (pprint *)
;;; (:form 2)
;;; meta-. translate-bodies
;;; (pprint (butlast * 3))
;;; meta-. translate-bodies

;;; Aha -- we have (UPDATE-NTH 0 MYARR MYPARENT).  Sigh, of course we can't
;;; have MYARR as a stobj there.  Yet we needed MYARR as a stobj for MYARRP.

;;; The thing is, of course update-nth can't work under the hood.  These are
;;; just logical definitions.  Maybe we can use :non-executable.  Then we don't
;;; need the stobj declaration.  How do I specify :non-executable?

;;; control-t g
;;; g declare
;;; g defun

;;; Oh wait -- I doubt that will work.  We need the function to be executable,
;;; albeit with an under-the-hood definition.  Ah, I see the use of non-exec on
;;; the creator.  I'll use non-exec -- not on the guard, which will need to be
;;; executed, but on the body.  Oops -- initially I'd removed stobj-xargs, but
;;; we still need that.  I also initially made an editing error, but compared
;;; with the original definition of defstobj-field-fns-axiomatic-defs to debug
;;; that.

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (stobj-formal ; !! maybe share with translate-stobj-type-to-guard [DONE]
              (and (not (eq type 'state))
                   (stobjp type t wrld)
                   type))
             (v-formal (or stobj-formal 'v))
             (stobj-xargs (and stobj-formal
                               `(:stobjs ,stobj-formal)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i ,v-formal ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t
                                             ,@stobj-xargs))
                             ,(if stobj-formal
                                  `(non-exec
                                    (update-nth-array ,n i ,v-formal ,var))
                                `(update-nth-array ,n i ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (eq type-term t)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t
                                            ,@stobj-xargs))
                             ,(if stobj-formal
                                  `(non-exec
                                    (update-nth ,n ,v-formal ,var))
                                `(update-nth ,n ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; Hurray!  Now Sol's defun goes through.

;;; So now I need to think about whether what I've done above is OK.
;;; One thing to look at is the body stored for update-parent-arr.

;;; :props update-parent-arr

;;; Aha -- I compare with old output from that :props, and I see a goofy guard:

#||
 (IF (MYARRP MYARR)
     (IF (MYARRP MYARR)
         (MYPARENTP MYPARENT)
         'NIL)
     'NIL)
||#

;;; Other than that, it all looks good.  I don't mind the RETURN-LAST in the
;;; UNNORMALIZED-BODY; it's not in the normalized body.

;;; Before thinking too much about correctness, let's fix the guard.

;;; Hmmm... it looks like it's conjoined the recognizer call from the stobj
;;; declaration with the explicit :guard.  Do we have that problem in general?
;;; Experiment:

#||
(defstobj st fld)
(defun foo (st)
  (declare (xargs :stobjs st
                  :guard (stp st)))
  (fld st))

ACL2 !>(guard 'foo nil (w state))
(IF (STP ST) (STP ST) 'NIL)
ACL2 !>(guard 'foo t (w state)) ; stobj-optp = t
'T
ACL2 !>
||#

;;; So yes, we have that problem in general, but it doesn't matter.  So it
;;; doesn't matter here either, most likely; let's check.

#||
ACL2 !>(guard 'update-parent-arr nil (w state))
(IF (MYARRP MYARR)
    (IF (MYARRP MYARR)
        (MYPARENTP MYPARENT)
        'NIL)
    'NIL)
ACL2 !>(guard 'update-parent-arr t (w state))
'T
ACL2 !>
||#

;;; Still, it's inelegant to lay down the extra conjunct; let's fix that.
;;; Hmmm, neither meta-. conjunct nor meta-. conjoin showed me a way to check if
;;; one term is a conjunct of another term.
;;; !! Reminder to self -- don't forget to test (and deal fully with) the case
;;; that we have an array of child stobjs. [DONE]

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (stobj-formal ; !! maybe share with translate-stobj-type-to-guard [DONE]
              (and (not (eq type 'state))
                   (stobjp type t wrld)
                   type))
             (v-formal (or stobj-formal 'v))
             (stobj-xargs (and stobj-formal
                               `(:stobjs ,stobj-formal)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i ,v-formal ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t
                                             ,@stobj-xargs))
                             ,(if stobj-formal
                                  `(non-exec
                                    (update-nth-array ,n i ,v-formal ,var))
                                `(update-nth-array ,n i ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (or (eq type-term t)
                                                     stobj-xargs)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t
                                            ,@stobj-xargs))
                             ,(if stobj-formal
                                  `(non-exec
                                    (update-nth ,n ,v-formal ,var))
                                `(update-nth ,n ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; That fixed the guard, and the other output from :props update-parent-arr is
;;; unchanged; also, Sol's example still goes through.  Let's deal now with
;;; arrays by considering this array version of Sol's example.

#||
(defstobj myarr (myarray-field :type (array (integer 0 *) (0)) :initially 0 :resizable t))

; Array of child stobjs this time:
(defstobj myparent (parent-arr :type (array myarr (8))))

(defun myparent-resize-array (size myparent)
  (declare (xargs :stobjs myparent :guard (natp size)
                  :guard-hints (("goal" :do-not-induct t))))
  (stobj-let ((myarr (parent-arri 0 myparent)))
             (myarr)
             (resize-myarray-field size myarr)
             myparent))
||#

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (array-etype (and arrayp (cadr type)))
             (stobj-formal ; !! maybe share with translate-stobj-type-to-guard [DONE]
              (cond (arrayp (and (not (eq array-etype 'state))
                                 (stobjp array-etype t wrld)
                                 array-etype))
                    (t (and (not (eq type 'state))
                            (stobjp type t wrld)
                            type))))
             (v-formal (or stobj-formal 'v))
             (stobj-xargs (and stobj-formal
                               `(:stobjs ,stobj-formal)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
             (,accessor-name (i ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var)))
                                             :verify-guards t))
                             (nth i (nth ,n ,var)))
             (,updater-name (i ,v-formal ,var)
                            (declare (xargs :guard
                                            (and (,top-recog ,var)
                                                 (integerp i)
                                                 (<= 0 i)
                                                 (< i (,length-name ,var))
                                                 ,@(if (eq array-etype-term
                                                           t)
                                                       nil
                                                     (list array-etype-term)))
                                            :verify-guards t
                                            ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth-array ,n i ,v-formal ,var))
                               `(update-nth-array ,n i ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (or (eq type-term t)
                                                     stobj-xargs)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t
                                            ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth ,n ,v-formal ,var))
                               `(update-nth ,n ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; Comparing :props looks good, except again for the guard.

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (array-etype (and arrayp (cadr type)))
             (stobj-formal ; !! maybe share with translate-stobj-type-to-guard [DONE]
              (cond (arrayp (and (not (eq array-etype 'state))
                                 (stobjp array-etype t wrld)
                                 array-etype))
                    (t (and (not (eq type 'state))
                            (stobjp type t wrld)
                            type))))
             (v-formal (or stobj-formal 'v))
             (stobj-xargs (and stobj-formal
                               `(:stobjs ,stobj-formal)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type v-formal wrld))))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype v-formal
                                                    wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
             (,accessor-name (i ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var)))
                                             :verify-guards t))
                             (nth i (nth ,n ,var)))
             (,updater-name (i ,v-formal ,var)
                            (declare
                             (xargs :guard
                                    (and (,top-recog ,var)
                                         (integerp i)
                                         (<= 0 i)
                                         (< i (,length-name ,var))

; We avoid laying down the stobj recognizer twice for a child stobj (although
; that would nevertheless be removed by the use of stobj-optp).

                                         ,@(and (not stobj-xargs)
                                                (if (eq array-etype-term
                                                        t)
                                                    nil
                                                  (list array-etype-term))))
                                    :verify-guards t
                                    ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth-array ,n i ,v-formal ,var))
                               `(update-nth-array ,n i ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (or (eq type-term t)

; We avoid laying down the stobj recognizer twice for a child stobj (although
; that would nevertheless be removed by the use of stobj-optp).

                                                     stobj-xargs)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t
                                            ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth ,n ,v-formal ,var))
                               `(update-nth ,n ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; Now all looks good.  Time to think about correctness and see what else
;;; might need to be done.

;;; First, though: it bothers me that we rely on the claim that (not
;;; stobj-xargs) implies that that type-term (or for arrays, array-etype-term)
;;; is the application of the child stobj recognizer to the child stobj.
;;; That's certainly the case, but if someday we change
;;; translate-stobj-type-to-guard, maybe it won't be.

;;; The first thing I noticed was that the comments in
;;; translate-stobj-type-to-guard need updating.  I think that the second
;;; paragraph as it was will become obsolete, because we now add stobj
;;; declarations -- it wasn't helpful anyhow, so I'll delete it without much
;;; thought.  The third paragraph didn't make sense, since after
;;;   (trace$ translate-declaration-to-guard)
;;;   (defstobj st (fld :type (satisfies acl2-numberp) :initially 0))
;;; we see that the first argument of translate-declaration-to-guard can be a
;;; term that has nothing to do with the main Lisp package.

(defun translate-stobj-type-to-guard (x var wrld)

; This function is a variant of translate-declaration-to-guard.  Like that
; function, x is an alleged type about the variable symbol var -- think
; (DECLARE (TYPE x ...)) -- and results in an UNTRANSLATED term about var if x
; is seen to be a valid type-spec for ACL2.  Unlike that function, here we
; allow x to be a stobj name, which may be used as a type in a field of another
; stobj (introduced after x).  We return nil if x is not either sort of valid
; type spec.

; Our intended use of this function is in generation of guards for field
; accessors and updaters, in the case of fields that are themselves stobjs.

  (or (translate-declaration-to-guard x var wrld)
      (let ((stobj-recog (and (not (eq x 'state))
                              (cadr

; Use stobjp below, not getprop, since we do not know that x is a symbol.

                               (stobjp x t wrld)))))
        (and stobj-recog
             (list stobj-recog var)))))

;;; But I wonder if I can eliminate this function completely, instead doing two
;;; separate computations each time so that I know when the predicate came from
;;; the stobj case.  Oh, look at this -- more to do!  (!! Don't forget. [DONE])

#||
~/acl2/acl2/books$ time fgrep --include='*.l*sp' -ri translate-stobj-type-to-guard .
./add-ons/hash-stobjs.lisp:                                (and ,(translate-stobj-type-to-guard
./add-ons/hash-stobjs.lisp:             (t (let ((type-term (translate-stobj-type-to-guard
./add-ons/hash-stobjs.lisp:                     (translate-stobj-type-to-guard type 'v wrld))))
./add-ons/hash-stobjs.lisp:                     (translate-stobj-type-to-guard array-etype 'v wrld))))

real	0m9.855s
user	0m6.017s
sys	0m1.137s
~/acl2/acl2/books$
||#

;;; Aha -- translate-stobj-type-to-guard is only used in
;;; defstobj-field-fns-axiomatic-defs and
;;; defstobj-component-recognizer-axiomatic-defs.  In the former, we can now
;;; simply use translate-declaration-to-guard instead, since that result is
;;; only used for the guards, where we no longer need the stobj recognizer
;;; since :stobjs is used explicitly for the child stobj.

;;; INSTALL
(defun translate-stobj-type-to-guard (x var wrld)

; This function is a variant of translate-declaration-to-guard.  Like that
; function, x is an alleged type about the variable symbol var -- think
; (DECLARE (TYPE x ...)) -- and results in an UNTRANSLATED term about var if x
; is seen to be a valid type-spec for ACL2.  Unlike that function, here we
; allow x to be a stobj name, which may be used as a type in a field of another
; stobj (introduced after x).  We return nil if x is not either sort of valid
; type spec.

; Our intended use of this function is in generation of guards for recognizers
; of stobj fields that may themselves be stobjs.  We do not use this however in
; accessors or updators, where translate-declaration-to-guard suffices: we do
; not want to generate a stobj recognizer since the child stobj is supplied
; explicitly using :stobjs.

  (or (translate-declaration-to-guard x var wrld)
      (let ((stobj-recog (and (not (eq x 'state))
                              (cadr

; Use stobjp below, not getprop, since we do not know that x is a symbol.

                               (stobjp x t wrld)))))
        (and stobj-recog
             (list stobj-recog var)))))

;;; INSTALL
(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (array-etype (and arrayp (cadr type)))
             (stobj-formal
              (cond (arrayp (and (not (eq array-etype 'state))
                                 (stobjp array-etype t wrld)
                                 array-etype))
                    (t (and (not (eq type 'state))
                            (stobjp type t wrld)
                            type))))
             (v-formal (or stobj-formal 'v))
             (stobj-xargs (and stobj-formal
                               `(:stobjs ,stobj-formal)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-declaration-to-guard type v-formal wrld))))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-declaration-to-guard array-etype v-formal
                                                     wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
             (,accessor-name (i ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var)))
                                             :verify-guards t))
                             (nth i (nth ,n ,var)))
             (,updater-name (i ,v-formal ,var)
                            (declare
                             (xargs :guard
                                    (and (,top-recog ,var)
                                         (integerp i)
                                         (<= 0 i)
                                         (< i (,length-name ,var))

; We avoid laying down the stobj recognizer twice for a child stobj (although
; that would nevertheless be removed by the use of stobj-optp).

                                         ,@(and (not stobj-xargs)
                                                (assert$
                                                 array-etype-term
                                                 (if (eq array-etype-term
                                                         t)
                                                     nil
                                                   (list array-etype-term)))))
                                    :verify-guards t
                                    ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth-array ,n i ,v-formal ,var))
                               `(update-nth-array ,n i ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (or (eq type-term t)

; We avoid laying down the stobj recognizer twice for a child stobj (although
; that would nevertheless be removed by the use of stobj-optp).

                                                     stobj-xargs)
                                                 `(,top-recog ,var)
                                               (assert$
                                                type-term
                                                `(and ,type-term
                                                      (,top-recog ,var))))
                                            :verify-guards t
                                            ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth ,n ,v-formal ,var))
                               `(update-nth ,n ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; Patch for books/add-ons/hash-stobjs.lisp:

;;; INSTALL in books/add-ons/hash-stobjs.lisp:
(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
;---<
             (hashp (and (consp type) (eq (car type) 'hash-table)))
             (hash-test (and hashp (cadr type)))
;   >---
             (array-etype (and arrayp (cadr type)))
             (stobj-formal
              (cond (arrayp (and (not (eq array-etype 'state))
                                 (stobjp array-etype t wrld)
                                 array-etype))
                    (t (and (not (eq type 'state))
                            (stobjp type t wrld)
                            type))))
             (v-formal (or stobj-formal 'v))
             (stobj-xargs (and stobj-formal
                               `(:stobjs ,stobj-formal)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
;---<
                   (not hashp)
;   >---
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-declaration-to-guard type v-formal wrld))))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-declaration-to-guard array-etype v-formal
                                                     wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable))
;---<
             (other (access defstobj-field-template
                            field-template
                            :other))
             (boundp-name (nth 0 other))
             (accessor?-name (nth 1 other))
             (remove-name (nth 2 other))
             (count-name (nth 3 other))
             (clear-name (nth 4 other))
             (init-name (nth 5 other))
;   >---
             )
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
             (,accessor-name (i ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var)))
                                             :verify-guards t))
                             (nth i (nth ,n ,var)))
             (,updater-name (i ,v-formal ,var)
                            (declare
                             (xargs :guard
                                    (and (,top-recog ,var)
                                         (integerp i)
                                         (<= 0 i)
                                         (< i (,length-name ,var))

; We avoid laying down the stobj recognizer twice for a child stobj (although
; that would nevertheless be removed by the use of stobj-optp).

                                         ,@(and (not stobj-xargs)
                                                (assert$
                                                 array-etype-term
                                                 (if (eq array-etype-term
                                                         t)
                                                     nil
                                                   (list array-etype-term)))))
                                    :verify-guards t
                                    ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth-array ,n i ,v-formal ,var))
                               `(update-nth-array ,n i ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
;---<
         (hashp
          (append
           `((,accessor-name
              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (cdr (hons-assoc-equal k (nth ,n ,var))))
             (,updater-name
              (k v ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (update-nth ,n (cons (cons k v) (nth ,n ,var)) ,var))
             (,boundp-name
              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (consp (hons-assoc-equal k (nth ,n ,var))))
             (,accessor?-name

              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (mv (,accessor-name k ,var)
                  (,boundp-name k ,var)))
             (,remove-name
              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (update-nth ,n (hons-remove-assoc k (nth ,n ,var)) ,var))
             (,count-name
              (,var)
              (declare (xargs :guard (,top-recog ,var)))
              (count-keys (nth ,n ,var)))
             (,clear-name
              (,var)
              (declare (xargs :guard (,top-recog ,var)))
              (update-nth ,n nil ,var))
             (,init-name
              (size rehash-size rehash-threshold ,var)
              (declare (xargs :guard (and (,top-recog ,var)
                                          (or (natp size)
                                              (not size))
                                          (or (and (rationalp rehash-size)
                                                   (< 1 rehash-size))
                                              (not rehash-size))
                                          (or (and (rationalp rehash-threshold)
                                                   (<= 0 rehash-threshold)
                                                   (<= rehash-threshold 1))
                                              (not rehash-threshold))))
                       (ignorable size rehash-size rehash-threshold))
              (update-nth ,n nil ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
;   >---
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (,v-formal ,var)
                            (declare (xargs :guard
                                            ,(if (or (eq type-term t)

; We avoid laying down the stobj recognizer twice for a child stobj (although
; that would nevertheless be removed by the use of stobj-optp).

                                                     stobj-xargs)
                                                 `(,top-recog ,var)
                                               (assert$
                                                type-term
                                                `(and ,type-term
                                                      (,top-recog ,var))))
                                            :verify-guards t
                                            ,@stobj-xargs))
                            ,(if stobj-formal
                                 `(non-exec
                                   (update-nth ,n ,v-formal ,var))
                               `(update-nth ,n ,v-formal ,var))))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;; OK, returning to the occurrence of 'V in chk-acceptable-defstobj: it
;;; suffices to add a comment.

;;; INSTALL
(defun chk-acceptable-defstobj (name args ctx wrld state)

; We check that (defstobj name . args) is well-formed and either
; signal an error or return nil.

  (cond
   ((not (symbolp name))
    (er soft ctx
        "The first argument of a DEFSTOBJ event must be a symbol.  Thus, ~x0 ~
          is ill-formed."
        (list* 'defstobj name args)))
   (t
    (mv-let
     (erp field-descriptors key-alist)
     (partition-rest-and-keyword-args args *defstobj-keywords*)
     (cond
      (erp
       (er soft ctx
           "The keyword arguments to the DEFSTOBJ event must appear after all ~
            field descriptors.  The allowed keyword arguments are ~&0, and ~
            these may not be duplicated, and must be followed by the ~
            corresponding value of the keyword argument.  Thus, ~x1 is ~
            ill-formed."
           *defstobj-keywords*
           (list* 'defstobj name args)))
      ((redundant-defstobjp name args wrld)
       (value 'redundant))
      (t
       (let ((renaming (cdr (assoc-eq :renaming key-alist)))
             (inline (cdr (assoc-eq :inline key-alist)))
             (congruent-to (cdr (assoc-eq :congruent-to key-alist)))
             (non-memoizable (cdr (assoc-eq :non-memoizable key-alist))))
         (cond
          ((not (booleanp inline))
           (er soft ctx
               "DEFSTOBJ requires the :INLINE keyword argument to have a ~
                Boolean value.  See :DOC defstobj."))
          ((not (booleanp non-memoizable))
           (er soft ctx
               "DEFSTOBJ requires the :NON-MEMOIZABLE keyword argument to ~
                have a Boolean value.  See :DOC defstobj."))
          ((and congruent-to
                (not (stobjp congruent-to t wrld)))
           (er soft ctx
               "The :CONGRUENT-TO field of a DEFSTOBJ must either be nil or ~
                the name of an existing stobj, but the value ~x0 is neither.  ~
                See :DOC defstobj."
               congruent-to))
          ((and congruent-to ; hence stobjp holds, hence symbolp holds
                (getpropc congruent-to 'absstobj-info nil wrld))
           (er soft ctx
               "The symbol ~x0 is the name of an abstract stobj in the ~
                current ACL2 world, so it is not legal for use as the ~
                :CONGRUENT-TO argument of DEFSTOBJ."
               congruent-to))
          ((and congruent-to
                (not (congruent-stobj-fields
                      field-descriptors
                      (car (old-defstobj-redundancy-bundle congruent-to
                                                           wrld)))))
           (er soft ctx
               "A non-nil :CONGRUENT-TO field of a DEFSTOBJ must be the name ~
                of a stobj that has the same shape as the proposed new stobj. ~
                ~ However, the proposed stobj named ~x0 does not have the ~
                same shape as the existing stobj named ~x1.  See :DOC ~
                defstobj."
               name congruent-to))
          ((and congruent-to
                (not (eq non-memoizable
                         (getpropc congruent-to 'non-memoizable nil wrld))))
           (er soft ctx
               "Congruent stobjs must agree on whether or not they are ~
                specified as :NON-MEMOIZABLE.  However, this fails for the ~
                proposed stobj, ~x0, which is specified as :CONGRUENT-TO the ~
                stobj ~x1, since ~x2 is specified with :NON-MEMOIZABLE T but ~
                ~x3 is not.  See :DOC defstobj."
               name
               congruent-to
               (if non-memoizable name congruent-to)
               (if non-memoizable congruent-to name)))
          (t
           (er-progn

; The defstobj name itself is not subject to renaming.  So we check it
; before we even bother to check the well-formedness of the renaming alist.

            (chk-all-but-new-name name ctx 'stobj wrld state)
            (cond ((or (eq name 'I)
                       (eq name 'V))

; Not only is 'v used in the logical definition of an updater when the field is
; not a child stobj (or array of such) -- also 'v is used in the raw definition
; of the updater in all cases.

                   (er soft ctx
                       "DEFSTOBJ does not allow single-threaded objects with ~
                        the names ~x0 or ~x1 because those symbols are used ~
                        as formals, along with the new stobj name itself, in ~
                        ``primitive'' stobj functions that will be defined."
                       'i 'v))
                  (t (value nil)))
            (chk-legal-defstobj-name name state)
            (cond ((not (doublet-style-symbol-to-symbol-alistp renaming))
                   (er soft ctx
                       "The :RENAMING argument to DEFSTOBJ must be an alist ~
                        containing elements of the form (sym sym), where each ~
                        element of such a doublet is a symbol. Your argument, ~
                        ~x0, is thus illegal."
                       renaming))
                  (t (value nil)))
            (er-let*
                ((wrld1 (chk-just-new-name name nil 'stobj nil ctx wrld state))
                 (wrld2 (chk-just-new-name (the-live-var name)
                                           nil 'stobj-live-var nil ctx wrld1
                                           state)))
              (chk-acceptable-defstobj1 name field-descriptors field-descriptors
                                        renaming non-memoizable
                                        ctx wrld2 state nil nil))))))))))))

;;;;;;; OLD definitions:

(defun translate-stobj-type-to-guard (x var wrld)

; This function is a variant of translate-declaration-to-guard.  Like that
; function, x is an alleged type about the variable symbol var -- think
; (DECLARE (TYPE x ...)) -- and results in an UNTRANSLATED term about var if x
; is seen to be a valid type-spec for ACL2.  Unlike that function, here we
; allow x to be a stobj name, which may be used as a type in a field of another
; stobj (introduced after x).  We return nil if x is not either sort of valid
; type spec.

; Our intended use of this function is in generation of guards for field
; accessors and updaters, in the case of fields that are themselves stobjs.
; When the defstobj is introduced, there are no stobj declarations, so the
; subfield's recognizer (stobj-recog, below) is called on an ordinary object.
; We started allowing such calls when translating for execution after
; Version_6.1; see translate11-call.

; Presumably the two arguments of OR below -- the guard corresponding to the
; type x, and the stobj recognizer term -- cannot both be non-nil, since types
; are in the main Lisp package and hence cannot be stobj names.  Nevertheless,
; in case we're wrong about that, we translate the type first, in order to
; avoid ambiguity in case there is a stobj previously defined locally in a book
; or encapsulate, for example (which would cause x to translate as a type in
; the first pass but to a type in the second pass).

  (or (translate-declaration-to-guard x var wrld)
      (let ((stobj-recog (and (not (eq x 'state))
                              (cadr

; Use stobjp below, not getprop, since we do not know that x is a symbol.

                               (stobjp x t wrld)))))
        (and stobj-recog
             (list stobj-recog var)))))

(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type 'v wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype 'v wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable)))
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i v ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t))
                             (update-nth-array ,n i v ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (v ,var)
                            (declare (xargs :guard
                                            ,(if (eq type-term t)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t))
                            (update-nth ,n v ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

(defun chk-acceptable-defstobj (name args ctx wrld state)

; We check that (defstobj name . args) is well-formed and either
; signal an error or return nil.

  (cond
   ((not (symbolp name))
    (er soft ctx
        "The first argument of a DEFSTOBJ event must be a symbol.  Thus, ~x0 ~
          is ill-formed."
        (list* 'defstobj name args)))
   (t
    (mv-let
     (erp field-descriptors key-alist)
     (partition-rest-and-keyword-args args *defstobj-keywords*)
     (cond
      (erp
       (er soft ctx
           "The keyword arguments to the DEFSTOBJ event must appear after all ~
            field descriptors.  The allowed keyword arguments are ~&0, and ~
            these may not be duplicated, and must be followed by the ~
            corresponding value of the keyword argument.  Thus, ~x1 is ~
            ill-formed."
           *defstobj-keywords*
           (list* 'defstobj name args)))
      ((redundant-defstobjp name args wrld)
       (value 'redundant))
      (t
       (let ((renaming (cdr (assoc-eq :renaming key-alist)))
             (inline (cdr (assoc-eq :inline key-alist)))
             (congruent-to (cdr (assoc-eq :congruent-to key-alist)))
             (non-memoizable (cdr (assoc-eq :non-memoizable key-alist))))
         (cond
          ((not (booleanp inline))
           (er soft ctx
               "DEFSTOBJ requires the :INLINE keyword argument to have a ~
                Boolean value.  See :DOC defstobj."))
          ((not (booleanp non-memoizable))
           (er soft ctx
               "DEFSTOBJ requires the :NON-MEMOIZABLE keyword argument to ~
                have a Boolean value.  See :DOC defstobj."))
          ((and congruent-to
                (not (stobjp congruent-to t wrld)))
           (er soft ctx
               "The :CONGRUENT-TO field of a DEFSTOBJ must either be nil or ~
                the name of an existing stobj, but the value ~x0 is neither.  ~
                See :DOC defstobj."
               congruent-to))
          ((and congruent-to ; hence stobjp holds, hence symbolp holds
                (getpropc congruent-to 'absstobj-info nil wrld))
           (er soft ctx
               "The symbol ~x0 is the name of an abstract stobj in the ~
                current ACL2 world, so it is not legal for use as the ~
                :CONGRUENT-TO argument of DEFSTOBJ."
               congruent-to))
          ((and congruent-to
                (not (congruent-stobj-fields
                      field-descriptors
                      (car (old-defstobj-redundancy-bundle congruent-to
                                                           wrld)))))
           (er soft ctx
               "A non-nil :CONGRUENT-TO field of a DEFSTOBJ must be the name ~
                of a stobj that has the same shape as the proposed new stobj. ~
                ~ However, the proposed stobj named ~x0 does not have the ~
                same shape as the existing stobj named ~x1.  See :DOC ~
                defstobj."
               name congruent-to))
          ((and congruent-to
                (not (eq non-memoizable
                         (getpropc congruent-to 'non-memoizable nil wrld))))
           (er soft ctx
               "Congruent stobjs must agree on whether or not they are ~
                specified as :NON-MEMOIZABLE.  However, this fails for the ~
                proposed stobj, ~x0, which is specified as :CONGRUENT-TO the ~
                stobj ~x1, since ~x2 is specified with :NON-MEMOIZABLE T but ~
                ~x3 is not.  See :DOC defstobj."
               name
               congruent-to
               (if non-memoizable name congruent-to)
               (if non-memoizable congruent-to name)))
          (t
           (er-progn

; The defstobj name itself is not subject to renaming.  So we check it
; before we even bother to check the well-formedness of the renaming alist.

            (chk-all-but-new-name name ctx 'stobj wrld state)
            (cond ((or (eq name 'I)
                       (eq name 'V))
                   (er soft ctx
                       "DEFSTOBJ does not allow single-threaded objects with ~
                        the names ~x0 or ~x1 because those symbols are used ~
                        as formals, along with the new stobj name itself, in ~
                        ``primitive'' stobj functions that will be defined."
                       'i 'v))
                  (t (value nil)))
            (chk-legal-defstobj-name name state)
            (cond ((not (doublet-style-symbol-to-symbol-alistp renaming))
                   (er soft ctx
                       "The :RENAMING argument to DEFSTOBJ must be an alist ~
                        containing elements of the form (sym sym), where each ~
                        element of such a doublet is a symbol. Your argument, ~
                        ~x0, is thus illegal."
                       renaming))
                  (t (value nil)))
            (er-let*
                ((wrld1 (chk-just-new-name name nil 'stobj nil ctx wrld state))
                 (wrld2 (chk-just-new-name (the-live-var name)
                                           nil 'stobj-live-var nil ctx wrld1
                                           state)))
              (chk-acceptable-defstobj1 name field-descriptors field-descriptors
                                        renaming non-memoizable
                                        ctx wrld2 state nil nil))))))))))))

;;;; From books/add-ons/hash-stobjs.lisp:
(defun defstobj-field-fns-axiomatic-defs (top-recog var n field-templates wrld)

; Wrld is normally a logical world, but it can be nil when calling this
; function from raw Lisp.

; Warning: Keep the formals in the definitions below in sync with corresponding
; formals defstobj-field-fns-raw-defs.  Otherwise trace$ may not work
; correctly; we saw such a problem in Version_5.0 for a resize function.

; Warning:  See the guard remarks in the Essay on Defstobj Definitions.

; We return a list of defs (see defstobj-axiomatic-defs) for all the accessors,
; updaters, and optionally, array resizing and length, of a single-threaded
; resource.

; Warning: Each updater definition should immediately follow the corresponding
; accessor definition, so that this is the case for the list of definitions
; returned by defstobj-axiomatic-defs.  That list of definitions becomes the
; 'stobj property laid down by defstobj-fn, and function
; chk-stobj-let/updaters1 assumes that it will find each updater definition in
; that property immediately after the corresponding accessor definition.

  (cond
   ((endp field-templates)
    nil)
   (t (let* ((field-template (car field-templates))
             (type (access defstobj-field-template
                           field-template
                           :type))
             (arrayp (and (consp type) (eq (car type) 'array)))
             (init0 (access defstobj-field-template
                            field-template
                            :init))
             (creator (get-stobj-creator (if arrayp (cadr type) type)
                                         wrld))
             (init (if creator
                       `(non-exec (,creator))
                     (kwote init0)))
;---<
             (hashp (and (consp type) (eq (car type) 'hash-table)))
             (hash-test (and hashp (cadr type)))
;   >---
             (type-term            ; used in guard
              (and (not arrayp)    ; else type-term is not used
;---<
                   (not hashp)
;   >---
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard type 'v wrld))))
             (array-etype (and arrayp (cadr type)))
             (array-etype-term     ; used in guard
              (and arrayp          ; else array-etype-term is not used
                   (if (null wrld) ; called from raw Lisp, so guard is ignored
                       t
                     (translate-stobj-type-to-guard array-etype 'v wrld))))
             (array-length (and arrayp (car (caddr type))))
             (accessor-name (access defstobj-field-template
                                    field-template
                                    :accessor-name))
             (updater-name (access defstobj-field-template
                                   field-template
                                   :updater-name))
             (length-name (access defstobj-field-template
                                  field-template
                                  :length-name))
             (resize-name (access defstobj-field-template
                                  field-template
                                  :resize-name))
             (resizable (access defstobj-field-template
                                field-template
                                :resizable))
;---<
             (other (access defstobj-field-template
                            field-template
                            :other))
             (boundp-name (nth 0 other))
             (accessor?-name (nth 1 other))
             (remove-name (nth 2 other))
             (count-name (nth 3 other))
             (clear-name (nth 4 other))
             (init-name (nth 5 other))
;   >---
             )
        (cond
         (arrayp
          (append
           `((,length-name (,var)
                           (declare (xargs :guard (,top-recog ,var)
                                           :verify-guards t)
                                    ,@(and (not resizable)
                                           `((ignore ,var))))
                           ,(if resizable
                                `(len (nth ,n ,var))
                              `,array-length))
             (,resize-name
              (i ,var)
              (declare (xargs :guard (,top-recog ,var)
                              :verify-guards t)
                       ,@(and (not resizable)
                              '((ignore i))))
              ,(if resizable
                   `(update-nth ,n
                                (resize-list (nth ,n ,var) i ,init)
                                ,var)
                 `(prog2$ (hard-error
                           ',resize-name
                           "The array field corresponding to accessor ~x0 of ~
                            stobj ~x1 was not declared :resizable t.  ~
                            Therefore, it is illegal to resize this array."
                           (list (cons #\0 ',accessor-name)
                                 (cons #\1 ',var)))
                          ,var)))
              (,accessor-name (i ,var)
                              (declare (xargs :guard
                                              (and (,top-recog ,var)
                                                   (integerp i)
                                                   (<= 0 i)
                                                   (< i (,length-name ,var)))
                                              :verify-guards t))
                              (nth i (nth ,n ,var)))
              (,updater-name (i v ,var)
                             (declare (xargs :guard
                                             (and (,top-recog ,var)
                                                  (integerp i)
                                                  (<= 0 i)
                                                  (< i (,length-name ,var))
                                                  ,@(if (eq array-etype-term
                                                            t)
                                                        nil
                                                      (list array-etype-term)))
                                             :verify-guards t))
                             (update-nth-array ,n i v ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
;---<
         (hashp
          (append
           `((,accessor-name
              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (cdr (hons-assoc-equal k (nth ,n ,var))))
             (,updater-name
              (k v ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (update-nth ,n (cons (cons k v) (nth ,n ,var)) ,var))
             (,boundp-name
              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (consp (hons-assoc-equal k (nth ,n ,var))))
             (,accessor?-name

              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (mv (,accessor-name k ,var)
                  (,boundp-name k ,var)))
             (,remove-name
              (k ,var)
              (declare (xargs :guard ,(if (eq hash-test 'eql)
                                          `(and (,top-recog ,var)
                                                (eqlablep k))
                                        `(,top-recog ,var))
                              :verify-guards t))
              (update-nth ,n (hons-remove-assoc k (nth ,n ,var)) ,var))
             (,count-name
              (,var)
              (declare (xargs :guard (,top-recog ,var)))
              (count-keys (nth ,n ,var)))
             (,clear-name
              (,var)
              (declare (xargs :guard (,top-recog ,var)))
              (update-nth ,n nil ,var))
             (,init-name
              (size rehash-size rehash-threshold ,var)
              (declare (xargs :guard (and (,top-recog ,var)
                                          (or (natp size)
                                              (not size))
                                          (or (and (rationalp rehash-size)
                                                   (< 1 rehash-size))
                                              (not rehash-size))
                                          (or (and (rationalp rehash-threshold)
                                                   (<= 0 rehash-threshold)
                                                   (<= rehash-threshold 1))
                                              (not rehash-threshold))))
                       (ignorable size rehash-size rehash-threshold))
              (update-nth ,n nil ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld)))
;   >---
         (t
          (append
           `((,accessor-name (,var)
                             (declare (xargs :guard (,top-recog ,var)
                                             :verify-guards t))
                             (nth ,n ,var))
             (,updater-name (v ,var)
                            (declare (xargs :guard
                                            ,(if (eq type-term t)
                                                 `(,top-recog ,var)
                                               `(and ,type-term
                                                     (,top-recog ,var)))
                                            :verify-guards t))
                            (update-nth ,n v ,var)))
           (defstobj-field-fns-axiomatic-defs
             top-recog var (+ n 1) (cdr field-templates) wrld))))))))

;;;;; FINALLY:

; Test Sol's example (both versions)
; :mini-proveall ; probably not necessary
; Document in note-7-4

;;; [!! Discovered later -- should have added to note-7-5 -- fixed that.]
(defxdoc note-7-4

; Total number of release note items: 22.

; Eliminate extra code for backward compatibility, now obsolete, in functions
; book-hash and cert-hash.

; Improved error messages from exceeding the call depth limit, in response to
; helpful feedback from Eric Smith.

; Here is a brief explanation of the soundness bug in functional instantiation,
; and its fix.  There was a "draconian" restriction until late 2006, which made
; it illegal for free variables in a functional subsitution to occur either in
; (a) the functionally-instantiated theorem or (b) the relevant
; functionally-instantiated constraints.  That restriction was eliminated in
; Version 3.0.2 (released December 2006); this was completely appropriate for,
; and motivated by, the desire to avoid the restriction for (a).  However, it
; was unsound to do this for (b): one is really functionally instantiating the
; universal closure of the constraints, so for example if function f is
; replaced by (lambda (u) (g u y)) in a constraint formula, say for example (p
; (f x) y), then logically we are substituting into the constraint (forall (x
; y) (p (f x) y)) to get the formula (forall (x y) (p (g x y) y)), which sadly
; is not a functional instance of the original constraint.  The new
; implementation automatically performs alpha renaming of the constraint to
; avoid such capture: in our example, it functionally instantiates the renamed
; constraint (forall (x y-renamed0) (p (f x) y-renamed0)) to obtain a suitable
; proof obligation of (p (g x y) y-renamed0).  A "[Capture]" warning is printed
; when this happens.

; Disabled executable-counterpart of function print-call-history; so now we
; don't get distracting output, for example from
; (thm (equal (print-call-history) 3)).

; Added a long comment, "Essay on Admitting a Model for Apply$ and the
; Functions that Use It", near the end of ACL2 source file apply-raw.lisp.
; This Essay is a step towards incorporating apply$, which supports a weak form
; of second-order programming and reasoning, into part of ACL2.  As mentioned
; there, [new] community books directory books/projects/apply-model/ contains
; examples that illustrate the Essay.

; Here is an example from Sol Swords showing the issue reported below for
; stobj-let and guard proofs.  Before the fix, the guard proof failed;
; afterwards, it succeeds.

;   (defstobj myarr (myarray-field :type (array (integer 0 *) (0))
;                                  :initially 0 :resizable t))
;
;   (defstobj myparent (parent-arr :type myarr))
;
;   (defun myparent-resize-array (size myparent)
;     (declare (xargs :stobjs myparent :guard (natp size)
;                     :guard-hints (("goal" :do-not-induct t))))
;     (stobj-let ((myarr (parent-arr myparent)))
;                (myarr)
;                (resize-myarray-field size myarr)
;                myparent))

  :parents (release-notes)
  :short "ACL2 Version  7.4 (March, 2017) Notes"
  :long "<p>NOTE!  New users can ignore these release notes, because the @(see
 documentation) has been updated to reflect all changes that are recorded
 here.</p>

 <p>Below we roughly organize the changes to ACL2 since Version 7.3 into the
 following categories of changes: existing features, new features, heuristic
 and efficiency improvements, bug fixes, changes at the system level, Emacs
 support, and experimental versions.  Each change is described in just one
 category, though of course many changes could be placed in more than one
 category.</p>

 <p>Note that only ACL2 system changes are listed below.  Changes to the @(see
 books) can be found by browsing the <a
 href='https://github.com/acl2/acl2/'>ACL2+Books GitHub repository</a>, in
 particular, the raw <a
 href='https://github.com/acl2/acl2/commits/master'>commit log</a>.  Also note
 that with each release, some built-in functions that were formerly in
 @(':')@(tsee program) mode are now @(see guard)-verified @(':')@(tsee logic)
 mode functions.</p>

 <h3>Changes to Existing Features</h3>

 <p>The utility @(tsee read-file-into-string) is now a macro, not a function.
 It no longer takes @('state') as an argument (though its expansion reads the
 state), and it has new keyword arguments that allow one to read successive
 segments of the file.  See @(see read-file-into-string).</p>

 <p>The test has been slightly relaxed for determining whether any supporting
 events are missing for @(tsee defabsstobj).  Thanks to Sol Swords for sending
 an example to illustrate the issue, together with analysis of the problem.</p>

 <h3>Heuristic and Efficiency Improvements</h3>

 <p>We now avoid the @(see invariant-risk) slowdown in the case of a
 @(':')@(tsee logic) mode function with guard @('t').  For relevant examples
 see @(see program-wrapper).</p>

 <p>A technical change in the handling of @(see type-set) computations can, in
 rare cases, result in better built-in @(see type-prescription) rules for
 functions.</p>

 <p>A small change in @(tsee defstobj) can lead to improved automation for some
 @(tsee guard) proofs involving @(tsee stobj-let).  Thanks to Sol Swords for
 sending an example (included in a Lisp comment in @('(deflabel note-7-4
 ...)')).</p>

 <h3>Bug Fixes</h3>

 <p>Fixed a soundness bug, introduced about 10 years ago, that could occur with
 @(':use') or @(':by') @(see lemma-instance)s when a variable occurs free in a
 lambda of a functional substitution.  For an explanation see a comment in
 @('(defxdoc note-7-4 ...)') in community book
 @('books/system/doc/acl2-doc.lisp'); also, a proof of @('nil') (in Version
 7.3) that exploits this bug can be found in a comment in the definition of
 function @('remove-capture-in-constraint-lst') in the ACL2 source code.  This
 may be the first time that a proved ``theorem'' in the @(see community-books)
 has been discovered to be a non-theorem (@('tr-to-array-of-array-to-tr') in
 @('books/projects/legacy-defrstobj/typed-records.lisp')).  Thanks to Sol
 Swords for pointing out that this directory of books isn't used elsewhere; we
 temporarily excluded it from the regression by adding a @('cert_pl_exclude')
 file, until Sol Swords repaired the proof a couple of days later.  Also thanks
 to Ruben Gamboa for fixing proofs in the @('nonstd/') books that failed after
 the bug fix.</p>

 <p>Fixed a bug in @(see proof-builder) commands @('x') and @('expand'), which
 usually gave a misleading error message when attempting to expand a
 constrained function.</p>

 <p>Fixed bugs in the tracking and disabling of @(see compound-recognizer)
 rules.  Thanks to Dmitry Nadezhin for pointing out this bug, indeed, sending
 an example and analyzing it to point to system function
 @('term-and-typ-to-lookup'), which was indeed the one that had the bug.</p>

 <p>Avoided a rare error in @(tsee hons-wash) that has happened when there is a
 second garbage collection after @('hons-wash') has commenced.  Thanks to
 Keshav Kini for telling us about the problem and providing a patch, which we
 used almost verbatim.</p>

 <h3>Changes at the System Level</h3>

 <p>An ACL2 customization file is no longer loaded by default (when it exists)
 when invoking `@('make')' from the top-level ACL2 directory.  This change is
 not necessary for `@('make regression')' because the build system for the
 books already takes care of it.  However we need this change for `@('make
 DOC')', at the least; thanks to Eric Smith for bringing our attention to a
 failure of `@('make DOC')'.</p>

 <p>(Really a books change, but affects building the manual:) By default, it is
 now an error if when saving the ACL2+books or ACL2-only manuals (web-based or
 text-based), a syntax error (marked by ``xdoc error'') is encountered.  See
 @(tsee xdoc::save) and @(tsee xdoc::save-rendered): both now have an error
 argument with default @('nil') so that by default, those new errors are not
 signaled; but in accordance with the change above, that new error argument is
 @('t') for the standard ways of building the ACL2+books and ACL2-only manuals.
 Also, two arguments have been eliminated for @('xdoc::save-rendered').</p>

 <p>An invocation of `@('make update')' to build an ACL2 executable initiates a
 build only when necessary (that is, when at least one source file or
 supporting file, like @('GNUmakefile'), has changed).  Thanks to Eric Smith
 for the suggestion.  Removed the following `@('make')' targets, which were
 suspect after this change (the first definitely no longer worked) and are
 unnecessary: @('saved_acl2'), @('saved_acl2p'), @('saved_acl2r'), and
 @('saved_acl2pr').  WARNING: @('make update') should be used with care, since
 it does not rebuild the desired executable when it already exists and is more
 recent than the sources.  For example, if you change Lisp implementations
 without changing the @('PREFIX') variable &mdash; perhaps even only changing
 the version of your Lisp &mdash; then use @('make') or @('make large'), not
 @('make update').</p>

 <p>New `@('make')' target @('clean-all') removes all generated files in the
 main ACL2 directory and its @('doc/') subdirectory, while existing target
 @('clean'), which has been renamed @('clean-lite'), cleans all of those except
 files @('*saved_acl2*') (and @('doc.lisp.backup')), essentially as before.
 Thanks to Eric Smith for suggestions leading to this change, and to Keshav
 Kini for suggesting consideration of target @('distclean') as an alias for
 @('clean-all'), which it now is.  The target @('clean') is deprecated but,
 consistent with prior usage, acts as @('clean-lite') for now.</p>

 <p>A successful build using `@('make')' shows only four lines for @(''fgrep -i
 compiler make.log')' instead of five.</p>

 <p>(GCL only) Installed a patch from Camm Maguire (GCL maintainer) to fix a
 GCL compiler bug.  Also turned off SGC to avoid an ``mprotect failure'' error
 (community book @('books/centaur/fty/tests/deftagsum-scale.lisp')).</p>

 <p>Fixed a bug that was causing an assertion failure, instead of a clean
 error, when attempting to apply @(':')@(tsee monitor) to undefined primitives,
 e.g., @(':monitor cons t').  Thanks to Keshav Kini for reporting this bug and
 sending us an initial fix.</p>

 <p>Fixed a bug in the @(see break-rewrite) commands @(':go!') and @(':go$'),
 which was causing an error message to be printed.</p>

 <h3>EMACS Support</h3>

 <p>Made minor @(see acl2-doc) improvements: fixed a bug in `@('u')' command's
 error message when already at the top topic; fixed a bug that was displaying
 the \"Parent list:\" field in reverse order; and eliminated the warning when
 loading a large manual (combined manual is at about 70M bytes now; new
 threshold is 80M bytes).</p>

 <p>Improved @(see acl2-doc) to work with other manuals; see @(see acl2-doc),
 specifically Section ``Selecting a Manual''.  Note that the command for
 initially bringing up the ACL2 User's Manual, instead of the ACL2+Books
 manual, has changed from @('(defvar *acl2-doc-top-default* 'TOP)') to
 @('(defvar *acl2-doc-manual-name* 'acl2-only)').</p>

 <p>Added new @(see acl2-doc) commands, `@('/')' and `@('W')', to go to an ACL2
 definition.  By default the search is through both the ACL2 sources and books.
 The search can be continued even outside acl2-doc by using a numeric prefix
 argument with @('control-t /'), which is just the `@('/')' command made
 available outside acl2-doc.  These two commands differ only in the defaults
 presented in the minibuffer.  The default for `@('/')' is based on the text at
 the cursor.  (This is much like the similar built-in Emacs command,
 @('meta-.'); indeed, `@('/')' is next to `@('.')' on standard keyboards.)  For
 `@('W')', the default is based on the page's topic name (as shown by the
 `@('w')' command).  See @(see acl2-doc) for details.</p>

 <p>Installed a patch for @('control-t control-e') from Keshav Kini, so that
 when there are more than two windows and one has the ACL2 shell buffer
 showing, then that one receives the form.  Consider for example windows A, B,
 and C, where the cursor is showing in A and the shell buffer is showing in C.
 Previously, the shell buffer could be placed in B to receive the form.  Now,
 the form will go into C.</p>

 <p>The @(':evisc-tuple') keyword for @(tsee trace$) was being ignored when
 @(':native t') was also supplied.  Since the use of @(':native t') sends the
 tracing command to raw Lisp, this wasn't really a bug, since no raw Lisp we
 know of supports the @(':evisc-tuple') keyword.  However, we have improved our
 custom modifications of the native @('trace') for host Lisps CCL and Allegro
 CL so that @(':evisc-tuple') is no longer ignored for these Lisps.</p>

 <h3>Experimental Versions</h3>

 <p>A new rewrite rule for ACL2(r), @('realp-implies-acl2-numberp'), is
 analogous to the existing rule @('rationalp-implies-acl2-numberp').  Thanks to
 Dmitry Nadezhin for suggesting the addition of this rule, which he observed is
 necessary for some ACL2(r) proofs.</p>

 ")

; Run "everything" regression
